#include "Storage.hpp"

namespace fw
{
    Storage::Storage()
    {
#ifdef ESP8266
        this->setup_eeprom();
#endif
    }

    Storage::~Storage()
    {
    }

    void Storage::setup_eeprom()
    {
#ifdef ESP8266
        EEPROM.begin(this->eeprom_size);
#endif
    }

    uint16_t Storage::eeprom_rule_position(uint8_t key)
    {
#ifdef ESP8266
        firewall_rule_t rule;
        uint8_t total_space_needed = 0;
        total_space_needed += sizeof(rule.source);
        total_space_needed += sizeof(rule.destination);
        total_space_needed += sizeof(rule.port_from);
        total_space_needed += sizeof(rule.port_to);
        total_space_needed += sizeof(rule.target);
        total_space_needed += sizeof(rule.protocol);
        // key-1 because key will be in range 1-255, but we need 1 less for multiplication
        return eeprom_rules_head + (key - 1) * total_space_needed;
#elif defined(ESP32)
        return 0;
#endif
    }

    uint8_t Storage::retrieve_settings_value(const char *key)
    {
#ifdef ESP32
        uint8_t value;

        this->memory.begin("settings", true);
        value = memory.getUChar(key, 0);
        this->memory.end();

        return value;
#elif defined(ESP8266)
        if (strncmp("amount_of_rules", key, 16) == 0)
        {
            uint8_t security_number = EEPROM.read(this->eeprom_settings_head);
            uint8_t amount_of_rules = EEPROM.read(this->eeprom_amout_of_rules_head);

            if (amount_of_rules > 50 || security_number != this->security_number)
                return 0;
            return amount_of_rules;
        }
        return 0;
#endif
    }

    void Storage::store_settings_value(const char *key, const uint8_t new_amount)
    {
#ifdef ESP32
        this->memory.begin("settings", false);
        this->memory.putUChar(key, new_amount);
        this->memory.end();
#elif defined(ESP8266)
        if (strncmp("amount_of_rules", key, 16) == 0)
        {
            EEPROM.write(this->eeprom_settings_head, this->security_number);
            EEPROM.write(this->eeprom_amout_of_rules_head, new_amount);
            EEPROM.commit();
        }
#endif
    }

    firewall_rule_t *Storage::retrieve_firewall_rule(const uint8_t key)
    {
        firewall_rule_t *rule_ptr = (firewall_rule_t *)malloc(sizeof(firewall_rule_t));
        rule_ptr->key = key;
#ifdef ESP32
        char rulename[9]; // fwRule99\n
        sprintf(rulename, "fwRule%i", key);

        this->memory.begin(rulename, true);
        strncpy(rule_ptr->source, this->memory.getString("source", "0.0.0.0").c_str(), sizeof(rule_ptr->source));
        strncpy(rule_ptr->destination, this->memory.getString("destination", "0.0.0.0").c_str(), sizeof(rule_ptr->source));
        rule_ptr->port_from = this->memory.getUChar("port_from", 0);
        rule_ptr->port_to = this->memory.getUChar("port_to", 0);
        rule_ptr->protocol = static_cast<firewall_protocol_t>(this->memory.getUChar("protocol", PROTOCOL_ALL));
        rule_ptr->target = static_cast<firewall_target_t>(this->memory.getUChar("target", TARGET_REJECT));
        this->memory.end();

#elif defined(ESP8266)
        uint16_t eespom_position = eeprom_rule_position(key);
        const char source[IPV4ADDRESS_LENGTH] = "";
        const char destination[IPV4ADDRESS_LENGTH] = "";

        EEPROM.get(eespom_position, source);
        strncpy(rule_ptr->source, source, sizeof(rule_ptr->source));
        eespom_position += sizeof(rule_ptr->source);

        EEPROM.get(eespom_position, destination);
        strncpy(rule_ptr->destination, destination, sizeof(rule_ptr->destination));
        eespom_position += sizeof(rule_ptr->destination);

        rule_ptr->port_from = EEPROM.read(eespom_position);
        eespom_position += sizeof(rule_ptr->port_from);

        rule_ptr->port_to = EEPROM.read(eespom_position);
        eespom_position += sizeof(rule_ptr->port_to);

        rule_ptr->protocol = static_cast<firewall_protocol_t>(EEPROM.read(eespom_position));
        eespom_position += sizeof(rule_ptr->protocol);

        rule_ptr->target = static_cast<firewall_target_t>(EEPROM.read(eespom_position));
#endif
        return rule_ptr;
    }

    void Storage::store_all_firewall_rules(firewall_rule_t *rule_head)
    {
#ifdef ESP32
        firewall_rule_t *temp = rule_head;
        while (temp != NULL)
        {
            store_firewall_rule(temp);
            temp = temp->next;
        }
#endif
    }

    void Storage::store_firewall_rule(firewall_rule_t *rule_ptr)
    {
#ifdef ESP32
        char rulename[9]; // fwRule99\n
        sprintf(rulename, "fwRule%i", rule_ptr->key);

        this->memory.begin(rulename, false);
        this->memory.putString("source", rule_ptr->source);
        this->memory.putString("destination", rule_ptr->destination);
        this->memory.putUChar("port_from", rule_ptr->port_from);
        this->memory.putUChar("port_to", rule_ptr->port_to);
        this->memory.putUChar("protocol", rule_ptr->protocol);
        this->memory.putUChar("target", rule_ptr->target);

        this->memory.end();
#elif defined(ESP8266)
        uint16_t eespom_position = eeprom_rule_position(rule_ptr->key);

        EEPROM.put(eespom_position, rule_ptr->source);
        eespom_position += sizeof(rule_ptr->source);
        EEPROM.put(eespom_position, rule_ptr->destination);
        eespom_position += sizeof(rule_ptr->destination);
        EEPROM.put(eespom_position, rule_ptr->port_from);
        eespom_position += sizeof(rule_ptr->port_from);
        EEPROM.put(eespom_position, rule_ptr->port_to);
        eespom_position += sizeof(rule_ptr->port_to);
        EEPROM.put(eespom_position, rule_ptr->protocol);
        eespom_position += sizeof(rule_ptr->protocol);
        EEPROM.put(eespom_position, rule_ptr->target);

        EEPROM.commit();
#endif
    }
}
